#!/usr/bin/env python3
"""
QoS Responsiveness Tests for Calico OpenStack Integration

This script tests the responsiveness of Calico's integration code in converting
QoS parameters from the Neutron API to the Calico WorkloadEndpoint API.

The tests verify that WorkloadEndpoint objects are correctly updated within
a few seconds when QoS policies are applied to networks and ports.
"""

import json
import logging
import os
import time
import unittest

import etcd3

import openstack


logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)

# openstack.enable_logging(debug=True, http_debug=True)


# Copied from
# https://github.com/tigera/calico-test/blob/bobcat/calicotest/common/testutils.py.
def retry_until_success(
    function,
    retries=10,
    wait_time=1,
    ex_class=None,
    log_exception=True,
    context_string="",
    function_args=None,
    function_kwargs=None,
):
    """
    Retries function until no exception is thrown. If exception continues,
    it is reraised.

    :param function: the function to be repeatedly called
    :param retries: the maximum number of times to retry the function.  A value
    of 0 will run the function once with no retries.
    :param wait_time: the time to wait between retries (in s)
    :param ex_class: The class of expected exceptions.
    :param log_exception: By default this function logs the exception if the
    function is still failing after max retries.   This log can sometimes be
    superfluous -- if e.g. the calling code is going to make a better log --
    and can be suppressed by setting this parameter to False.
    :param context_string: A string used to flag the context of a specific call
    in logs.
    :param function_args: A list of arguments to pass to function
    :param function_kwargs: A dictionary of keyword arguments to pass to
                            function
    :returns: the value returned by function
    """
    if function_args is None:
        function_args = []
    if function_kwargs is None:
        function_kwargs = {}
    for retry in range(int(retries) + 1):
        try:
            result = function(*function_args, **function_kwargs)
        except Exception as e:
            if ex_class and e.__class__ is not ex_class:
                logger.exception("Hit unexpected exception in function - not retrying.")
                raise
            if retry < retries:
                logger.debug("Hit exception in function - retrying: %s", e)
                time.sleep(wait_time)
            else:
                if log_exception:
                    logger.exception(
                        "Function %s did not succeed before timeout.", function
                    )
                raise
        else:
            # Successfully ran the function
            return result


class QoSResponsivenessTest(unittest.TestCase):
    """Test suite for QoS responsiveness in Calico OpenStack integration."""

    @classmethod
    def setUpClass(cls):
        cls.next_subnet_byte = 0

        # Set up OpenStack connection
        cls.conn = openstack.connection.Connection(
            auth_url=os.environ.get("OS_AUTH_URL", "http://localhost/identity"),
            project_name=os.environ.get("OS_PROJECT_NAME", "admin"),
            username=os.environ.get("OS_USERNAME", "admin"),
            password=os.environ.get("OS_PASSWORD", "015133ea2bdc46ed434c"),
            region_name=os.environ.get("OS_REGION_NAME", "RegionOne"),
            project_domain_id=os.environ.get("OS_PROJECT_DOMAIN_ID", "default"),
            user_domain_id=os.environ.get("OS_USER_DOMAIN_ID", "default"),
            identity_api_version=3,
        )
        logger.info("OpenStack connection established")

        # Set up etcd client for Calico datastore access
        etcd_host = os.environ.get("ETCD_HOST", "localhost")
        etcd_port = int(os.environ.get("ETCD_PORT", "2379"))
        cls.etcd_client = etcd3.client(host=etcd_host, port=etcd_port)
        logger.info(f"etcd3 client established: {etcd_host}:{etcd_port}")
        status = cls.etcd_client.status()
        logger.info(f"status.version = {status.version}")
        logger.info(f"status.db_size = {status.db_size}")
        logger.info(f"status.leader = {status.leader}")
        logger.info(f"status.raft_index = {status.raft_index}")
        logger.info(f"status.raft_term = {status.raft_term}")

        # Define possible QoS rules that we will use in this test.
        cls.qos_rules = [
            {
                "rule": {
                    "type": "bandwidth_limit",
                    "max_kbps": 10200,
                    "max_burst_kbps": 20300,
                    "direction": "egress",
                },
                "controls": {
                    "egressBandwidth": 10200000,
                    "egressPeakrate": 20300000,
                    "egressBurst": 4294967296,
                },
            },
            {
                "rule": {
                    "type": "bandwidth_limit",
                    "max_kbps": 30400,
                    "max_burst_kbps": 40500,
                    "direction": "ingress",
                },
                "controls": {
                    "ingressBandwidth": 30400000,
                    "ingressPeakrate": 40500000,
                    "ingressBurst": 4294967296,
                },
            },
        ]
        # The Yoga version of openstacksdk does not support creating packet
        # rate limit rules.  Caracal and onwards do support this.
        if hasattr(cls.conn.network, "create_qos_packet_rate_limit_rule"):
            logger.info("openstacksdk can create packet rate limit rules")
            cls.qos_rules.extend(
                [
                    {
                        "rule": {
                            "type": "packet_rate_limit",
                            "max_kpps": 6,
                            "direction": "egress",
                        },
                        "controls": {
                            "egressPacketRate": 6000,
                            "egressPacketBurst": 5,
                        },
                    },
                    {
                        "rule": {
                            "type": "packet_rate_limit",
                            "max_kpps": 7,
                            "direction": "ingress",
                        },
                        "controls": {
                            "ingressPacketRate": 7000,
                            "ingressPacketBurst": 5,
                        },
                    },
                ]
            )

        # Define possible combination sets of those rules.  If there are N
        # possible rules, there are 2**N possible combinations of them, formed
        # by each set including or not including each rule.
        cls.qos_rule_sets = []
        for i in range(2 ** len(cls.qos_rules)):
            rule_set = []
            for j in range(len(cls.qos_rules)):
                if i & (2**j) != 0:
                    rule_set.append(cls.qos_rules[j])
            cls.qos_rule_sets.append(rule_set)

    def test_qos_policy_id(self):
        # Test that a WorkloadEndpoint always gets the correct QoS settings as
        # we change the qos_policy_id on the corresponding port and on that
        # port's network.
        #
        # In this test, the rule set for a given qos_policy_id stays the same.

        # Form a set of states, in which transitioning between two of that
        # states means changing the qos_policy_id on the port and/or on the
        # network - including the possibilities that qos_policy_id might be not
        # set on either of those.
        policy_id_to_rule_set = {
            "A": self.qos_rule_sets[1],
            "B": self.qos_rule_sets[2],
            "C": self.qos_rule_sets[3],
        }
        states = []
        for net_qos_name in [None, "A", "B"]:
            for port_qos_name in [None, "B", "C"]:
                state = {
                    "net_qos_name": net_qos_name,
                    "port_qos_name": port_qos_name,
                }
                if port_qos_name is not None:
                    state["port_qos_rules"] = policy_id_to_rule_set[port_qos_name]
                else:
                    state["port_qos_rules"] = []
                if net_qos_name is not None:
                    state["net_qos_rules"] = policy_id_to_rule_set[net_qos_name]
                else:
                    state["net_qos_rules"] = []
                states.append(state)

        # Test transitions between those states.
        self._test_transitions(states)

    def test_qos_rule_change(self):
        # Test that a WorkloadEndpoint always gets the correct QoS settings as
        # we change the QoS rules for the qos_policy_id on the corresponding
        # port and on that port's network.
        #
        # In this test, the qos_policy_id values themselves stay the same.

        # Form a set of states, in which transitioning between two of that
        # states means changing the QoS rules for the effective qos_policy_id.
        state_base = {
            "net_qos_name": None,
            "port_qos_name": "D",
        }
        states = []
        for rules in self.qos_rule_sets:
            state = state_base.copy()
            state["port_qos_rules"] = rules
            state["net_qos_rules"] = []
            states.append(state)

        # Test transitions between those states.
        self._test_transitions(states)

    def _test_transitions(self, states):
        # Calculate the possible transitions from each state, such that only
        # one thing is changing in each transition.
        transitions_remaining = 0
        for i in range(len(states)):
            states[i]["transitions"] = []
            states[i]["transitions_covered"] = []
            for j in range(len(states)):
                if i == j:
                    continue
                if self._only_one_change(states[i], states[j]):
                    states[i]["transitions"].append(j)
                    transitions_remaining += 1

        # Calculate a set of state sequences that will cover all of the
        # possible single-change transitions.
        sequences = []
        while transitions_remaining:
            logger.info(
                "%d transitions remaining, start new sequence", transitions_remaining
            )
            sequence = []
            # Find an initial state.
            for i in range(len(states)):
                if len(states[i]["transitions_covered"]) < len(
                    states[i]["transitions"]
                ):
                    sequence.append(i)
                    break
            making_progress = True
            while making_progress:
                i = sequence[-1]
                making_progress = False
                for j in states[i]["transitions"]:
                    if j in states[i]["transitions_covered"]:
                        continue
                    state_transitions_remaining = len(states[i]["transitions"]) - len(
                        states[i]["transitions_covered"]
                    )
                    if (
                        state_transitions_remaining > 1
                        and len(sequence) > 1
                        and j == sequence[-2]
                    ):
                        continue
                    sequence.append(j)
                    states[i]["transitions_covered"].append(j)
                    transitions_remaining -= 1
                    making_progress = True
                    break
            logger.info(
                "Calculated sequence with length %d = %r", len(sequence), sequence
            )
            sequences.append(sequence)

        # Test each sequence in turn.
        for sequence in sequences:
            self._test_sequence([states[i] for i in sequence])

    def _only_one_change(self, a, b):
        changes = []
        if a["net_qos_name"] != b["net_qos_name"]:
            changes.append(f"change net_qos_name to {b['net_qos_name']}")
        if a["port_qos_name"] != b["port_qos_name"]:
            changes.append(f"change port_qos_name to {b['port_qos_name']}")
        if len(changes) > 1:
            return False
        # Only compare the rules when IDs are not changing.
        if len(changes) == 0:
            if b["port_qos_name"] is not None:
                relevant_rules = "port_qos_rules"
            else:
                relevant_rules = "net_qos_rules"
            a_rules = a[relevant_rules]
            b_rules = b[relevant_rules]
            for r in a_rules:
                if r not in b_rules:
                    changes.append(f"remove rule {r}")
            for r in b_rules:
                if r not in a_rules:
                    changes.append(f"add rule {r}")
        if len(changes) > 1:
            return False
        if len(changes) == 0:
            return False
        return changes[0]

    def _test_sequence(self, sequence):
        current = None
        for nxt in sequence:
            if current is None:
                port_id = self._create_initial_state(nxt)
            else:
                self._apply_change(current, nxt)
                self._verify_wep_qos(port_id, nxt)
            current = nxt
        self.tearDown()

    def _create_initial_state(self, state):
        logger.info(f"Create initial state -> {state}")

        # Create the QoS rules and policies that we need.
        if state["port_qos_name"] is not None:
            port_qos_id = self._ensure_qos_policy(
                state["port_qos_name"],
                state["port_qos_rules"],
            )
        else:
            port_qos_id = None

        if state["net_qos_name"] is not None:
            net_qos_id = self._ensure_qos_policy(
                state["net_qos_name"],
                state["net_qos_rules"],
            )
        else:
            net_qos_id = None

        # Create the network and subnet.
        network, subnet = self.create_test_network("test-network", net_qos_id)

        # Create the VM.
        cirros = [i for i in self.conn.image.images() if i.name.startswith("cirros")][0]
        tiny = [i for i in self.conn.compute.flavors() if "tiny" in i.name][0]
        vm = self.conn.compute.create_server(
            name="test-vm",
            image_id=cirros.id,
            flavor_id=tiny.id,
            networks=[{"uuid": network.id}],
        )
        vm = self.conn.compute.wait_for_server(vm)
        port = list(self.conn.network.ports(device_id=vm.id))[0]

        # Maybe set port-level QoS.
        if port_qos_id is not None:
            self.conn.network.update_port(port.id, qos_policy_id=port_qos_id)

        return port.id

    def _ensure_qos_policy(self, name, rules):
        full_name = "test-qos-policy" + name
        qos_policy = self.conn.network.find_qos_policy(full_name)
        if qos_policy is None:
            qos_policy = self.conn.network.create_qos_policy(name=full_name)

        significant_keys = [
            "type",
            "direction",
            "max_burst_kbps",
            "max_kbps",
            "max_kpps",
        ]
        existing_rules = [
            {k: v for k, v in r.items() if k in significant_keys}
            for r in qos_policy.rules
        ]
        desired_rules = [
            {k: v for k, v in r["rule"].items() if k in significant_keys} for r in rules
        ]

        for i, r in enumerate(existing_rules):
            if r not in desired_rules:
                logger.info(f"Delete rule {r} for policy {name}")
                if r["type"] == "bandwidth_limit":
                    self.conn.network.delete_qos_bandwidth_limit_rule(
                        qos_policy.rules[i]["id"], qos_policy.id
                    )
                elif r["type"] == "packet_rate_limit":
                    self.conn.network.delete_qos_packet_rate_limit_rule(
                        qos_policy.rules[i]["id"], qos_policy.id
                    )

        for r in desired_rules:
            if r not in existing_rules:
                logger.info(f"Add rule {r} for policy {name}")
                if r["type"] == "bandwidth_limit":
                    r2 = r.copy()
                    del r2["type"]
                    self.conn.network.create_qos_bandwidth_limit_rule(
                        qos_policy.id, **r2
                    )
                elif r["type"] == "packet_rate_limit":
                    r2 = r.copy()
                    del r2["type"]
                    self.conn.network.create_qos_packet_rate_limit_rule(
                        qos_policy.id, **r2
                    )

        return qos_policy.id

    def create_test_network(self, name, qos_policy_id=None):
        """Create a test network and subnet."""
        network_args = {
            "name": name,
            "shared": True,
            "provider:network_type": "local",
        }
        if qos_policy_id:
            network_args["qos_policy_id"] = qos_policy_id

        network = self.conn.network.create_network(**network_args)

        subnet = self.conn.network.create_subnet(
            name=f"{name}-subnet",
            network_id=network.id,
            cidr="10.63.%d.0/24" % self.next_subnet_byte,
            ip_version=4,
            enable_dhcp=True,
        )
        logger.info(
            f"Created network: {name}"
            f" {'with QoS policy' if qos_policy_id else 'without QoS policy'}"
        )
        self.next_subnet_byte += 1
        self.assertLess(self.next_subnet_byte, 256)

        return network, subnet

    def _apply_change(self, current, nxt):
        change = self._only_one_change(current, nxt)
        logger.info(f" {change} -> {nxt}")
        if change.startswith("change net_qos_name"):
            if nxt["net_qos_name"] is not None:
                net_qos_id = self._ensure_qos_policy(
                    nxt["net_qos_name"], nxt["net_qos_rules"]
                )
            else:
                net_qos_id = None
            network = self.conn.network.find_network("test-network")
            self.conn.network.update_network(network.id, qos_policy_id=net_qos_id)
        elif change.startswith("change port_qos_name"):
            if nxt["port_qos_name"] is not None:
                port_qos_id = self._ensure_qos_policy(
                    nxt["port_qos_name"], nxt["port_qos_rules"]
                )
            else:
                port_qos_id = None
            vm = self.conn.compute.find_server("test-vm")
            port = list(self.conn.network.ports(device_id=vm.id))[0]
            self.conn.network.update_port(port.id, qos_policy_id=port_qos_id)
        elif change.startswith("remove rule") or change.startswith("add rule"):
            if nxt["net_qos_name"] is not None:
                self._ensure_qos_policy(nxt["net_qos_name"], nxt["net_qos_rules"])
            if nxt["port_qos_name"] is not None:
                self._ensure_qos_policy(nxt["port_qos_name"], nxt["port_qos_rules"])

    def _verify_wep_qos(self, port_id, state):
        logger.info(f"Verify WEP QoS for state {state}")
        expected_qos = {}
        if state["port_qos_name"] is not None:
            for r in state["port_qos_rules"]:
                expected_qos.update(r["controls"])
        elif state["net_qos_name"] is not None:
            for r in state["net_qos_rules"]:
                expected_qos.update(r["controls"])
        logger.info(f"Expected QoS is {expected_qos}")
        retry_until_success(
            self._assert_wep_qos,
            function_args=(port_id, expected_qos),
        )

    def _assert_wep_qos(self, port_id, expected_qos):
        """
        Assert that WorkloadEndpoint QoS controls are as expected.

        Args:
            port_id: Neutron port ID
            expected_qos: Expected QoS controls dictionary
        """
        wep = None
        for value, metadata in self.etcd_client.get_prefix(
            "/calico/resources/v3/projectcalico.org/workloadendpoints/"
        ):
            logger.info(f"Metadata = {metadata}")
            if port_id.replace("-", "--") in metadata.key.decode():
                wep = json.loads(value.decode())
                break

        logger.info(f"WEP for port {port_id} is {wep}")
        self.assertIsNotNone(wep)
        self.assertIn("spec", wep)
        if expected_qos:
            self.assertIn("qosControls", wep["spec"])
            qos_controls = wep["spec"]["qosControls"]
            self.assertDictEqual(qos_controls, expected_qos)
        else:
            self.assertNotIn("qosControls", wep["spec"])

    def setUp(self):
        self.tearDown()

    def tearDown(self):
        logger.info("Delete VM")
        vm = self.conn.compute.find_server("test-vm")
        if vm is not None:
            self.conn.compute.delete_server(vm.id)
            retry_until_success(
                lambda: self.assertIsNone(self.conn.compute.find_server("test-vm")),
                wait_time=5,
            )

        logger.info("Delete subnet")
        subnet = self.conn.network.find_subnet("test-network-subnet")
        if subnet is not None:
            self.conn.network.delete_subnet(subnet.id)

        logger.info("Delete network")
        network = self.conn.network.find_network("test-network")
        if network is not None:
            self.conn.network.delete_network(network.id)

        logger.info("Delete QoS")
        for name in ["A", "B", "C", "D"]:
            full_name = "test-qos-policy" + name
            policy = self.conn.network.find_qos_policy(full_name)
            if policy:
                self.conn.network.delete_qos_policy(policy.id)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
networking_calico.plugins.ml2.drivers.calico.test.test_monitor_thread

Unit tests for the thread that monitors the periodic resync thread.
"""
from datetime import datetime, timedelta
import mock
import unittest

import networking_calico.plugins.ml2.drivers.calico.test.lib as lib
from networking_calico.plugins.ml2.drivers.calico import mech_calico


INITIAL_EPOCH = 0
TEST_MAX_INTERVAL = 30


class TestResyncMonitorThread(lib.Lib, unittest.TestCase):
    """Tests for the driver's resync monitor thread logic."""

    def setUp(self):
        super(TestResyncMonitorThread, self).setUp()

        # thread logic mocks
        self.driver.elector = mock.Mock()
        self.sleep_patcher = mock.patch("eventlet.sleep")
        self.mock_sleep = self.sleep_patcher.start()

        # log mocks
        self.log_error = mock.patch.object(mech_calico.LOG, "error").start()
        self.log_info = mock.patch.object(mech_calico.LOG, "info").start()
        self.log_debug = mock.patch.object(mech_calico.LOG, "debug").start()

        # resync mocks
        self.driver.subnet_syncer = mock.Mock()
        self.driver.policy_syncer = mock.Mock()
        self.driver.endpoint_syncer = mock.Mock()
        self.driver.provide_felix_config = mock.Mock()

    def tearDown(self):
        self.sleep_patcher.stop()
        super(TestResyncMonitorThread, self).tearDown()

    def simulate_epoch_progression(self, expected_sleep_time=None):
        def increment_epoch(actual_sleep_time):
            if expected_sleep_time is not None:
                assert expected_sleep_time == actual_sleep_time
            self.driver._epoch += 1

        return increment_epoch

    def test_monitor_does_nothing_when_not_master(self):
        """Test that a driver that is not master does not monitor."""
        self.driver.elector.master.return_value = False
        self.mock_sleep.side_effect = self.simulate_epoch_progression()

        self.driver.resync_monitor_thread(INITIAL_EPOCH)

        self.log_debug.assert_called_once_with("I am not master")
        self.log_error.assert_not_called()

    def test_monitor_logs_error_when_over_max(self):
        """Test that an error is logged when interval surpasses maximum."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True
        fake_resync_time = datetime.now() - timedelta(seconds=TEST_MAX_INTERVAL + 1)
        self.driver.last_resync_time = fake_resync_time
        self.mock_sleep.side_effect = self.simulate_epoch_progression()

        self.driver.resync_monitor_thread(INITIAL_EPOCH)

        self.log_error.assert_called_once()
        self.assertIn(
            "The time since the last resync completion has surpassed",
            self.log_error.call_args[0][0],
        )

    def test_monitor_no_error_if_interval_under_max(self):
        """If interval is below max, no error should be logged."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True
        self.mock_sleep.side_effect = self.simulate_epoch_progression()

        self.driver.resync_monitor_thread(INITIAL_EPOCH)

        self.log_error.assert_not_called()

    def test_monitor_exception_stops_elector(self):
        """On unexpected exception, elector.stop() must be called."""
        self.driver.elector.master.return_value = True

        with mock.patch.object(self.driver, "elector") as mock_elector:
            mock_elector.master.side_effect = Exception("Test exception")

            with self.assertRaises(Exception):
                self.driver.resync_monitor_thread(INITIAL_EPOCH)

            mock_elector.stop.assert_called_once()

    def test_resync_resets_time(self):
        """Test that resync resets current interval duration to below max."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True
        fake_resync_time = datetime.now() - timedelta(seconds=TEST_MAX_INTERVAL + 1)
        self.driver.last_resync_time = fake_resync_time

        self.mock_sleep.side_effect = self.simulate_epoch_progression()
        self.driver.resync_monitor_thread(INITIAL_EPOCH)

        self.log_error.assert_called_once()
        self.assertIn(
            "The time since the last resync completion has surpassed",
            self.log_error.call_args[0][0],
        )

        # Resync
        self.mock_sleep.side_effect = self.simulate_epoch_progression()
        self.driver.periodic_resync_thread(INITIAL_EPOCH + 1)

        self.mock_sleep.side_effect = self.simulate_epoch_progression()
        self.driver.resync_monitor_thread(INITIAL_EPOCH + 2)

        self.log_error.assert_called_once()

    def test_errors_continue_to_log(self):
        """Test that errors continue logging if resync does not occur."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True
        fake_resync_time = datetime.now() - timedelta(seconds=TEST_MAX_INTERVAL + 1)
        self.driver.last_resync_time = fake_resync_time

        self.mock_sleep.side_effect = self.simulate_epoch_progression()
        self.driver.resync_monitor_thread(INITIAL_EPOCH)

        self.log_error.assert_called_once()
        self.assertIn(
            "The time since the last resync completion has surpassed",
            self.log_error.call_args[0][0],
        )

        self.mock_sleep.side_effect = self.simulate_epoch_progression()
        self.driver.resync_monitor_thread(INITIAL_EPOCH + 1)

        self.assertEqual(self.log_error.call_count, 2)
        self.assertIn(
            "The time since the last resync completion has surpassed",
            self.log_error.call_args[0][0],
        )

    @mock.patch("networking_calico.plugins.ml2.drivers.calico.mech_calico.datetime")
    def test_sleep_time_logic_before_deadline(self, mock_datetime):
        """Test that we sleep until deadline if there is time left."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True

        curr_time = datetime.now()
        self.driver.last_resync_time = curr_time
        expected_sleep_time = TEST_MAX_INTERVAL
        mock_datetime.now.return_value = curr_time

        self.mock_sleep.side_effect = self.simulate_epoch_progression(
            expected_sleep_time
        )
        self.driver.resync_monitor_thread(INITIAL_EPOCH)

    def test_sleep_time_logic_after_deadline(self):
        """Test that we poll if the deadline has passed."""
        lib.m_oslo_config.cfg.CONF.calico.resync_max_interval_secs = TEST_MAX_INTERVAL
        self.driver.elector.master.return_value = True

        fake_resync_time = datetime.now() - timedelta(seconds=TEST_MAX_INTERVAL + 1)
        self.driver.last_resync_time = fake_resync_time
        expected_sleep_time = TEST_MAX_INTERVAL / 5

        self.mock_sleep.side_effect = self.simulate_epoch_progression(
            expected_sleep_time
        )
        self.driver.resync_monitor_thread(INITIAL_EPOCH)



# Copyright (c) 2022-2023 Tigera, Inc. All rights reserved.

import logging
import re

from tests.k8st.test_base import Container, Pod, TestBase
from tests.k8st.utils.utils import DiagsCollector, calicoctl, kubectl, run, retry_until_success, node_info, start_external_node_with_bgp

_log = logging.getLogger(__name__)

bird_conf = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 64512;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

# ------------- Node-to-node mesh -------------
protocol bgp Mesh_with_node_1 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}
"""

class TestBGPFilter(TestBase):
    def setUp(self):
        super(TestBGPFilter, self).setUp()

        # Create test namespace
        self.ns = "bgpfilter-test"
        self.create_namespace(self.ns)

        self.nodes, self.ips, self.ip6s = node_info()
        self.egress_node = self.nodes[1]
        self.egress_node_ip = self.ips[1]
        self.egress_node_ip6 = self.ip6s[1]
        self.external_node_ip = start_external_node_with_bgp(
            "kube-node-extra",
            bird_peer_config=self._get_bird_conf(),
        )
        self.external_node_ip6 = start_external_node_with_bgp(
            "kube-node-extra-v6",
            bird6_peer_config=self._get_bird_conf(ipv6=True),
        )

        kubectl("label node %s egress=true --overwrite" % self.egress_node)

        self.egress_calico_pod = self.get_calico_node_pod(self.egress_node)

        # Establish BGPPeer from cluster nodes to node-extra
        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: node-extra.peer
spec:
  peerIP: %s
  asNumber: 64512
  nodeSelector: "egress == 'true'"
EOF
""" % self.external_node_ip)

        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: node-extra-v6.peer
spec:
  peerIP: %s
  asNumber: 64512
  nodeSelector: "egress == 'true'"
EOF
""" % self.external_node_ip6)

    def tearDown(self):
        # Ensure that we always attempt to run the base class tearDown, storing the
        # exception if it fails and re-raising it after we've done our own teardown.
        teardown_exception = None
        try:
            super(TestBGPFilter, self).tearDown()
        except Exception as e:
            # Log the error, but don't block further teardown steps.
            _log.error("Exception during tearDown: %s", e)
            teardown_exception = e

        self.delete_and_confirm(self.ns, "ns")

        try:
            # Delete the extra node.
            run("docker rm -f kube-node-extra")
        except subprocess.CalledProcessError:
            pass

        try:
            # Delete the extra node.
            run("docker rm -f kube-node-extra-v6")
        except subprocess.CalledProcessError:
            pass

        # Delete BGPPeers.
        kubectl("delete bgppeer node-extra.peer", allow_fail=True)
        kubectl("delete bgppeer node-extra-v6.peer", allow_fail=True)

        if teardown_exception:
            raise teardown_exception

    def _get_bird_conf(self, ipv6=False):
        if ipv6:
            return bird_conf % (self.egress_node_ip6)

        return bird_conf % (self.egress_node_ip)

    def _check_route_in_cluster_bird(self, calicoPod, route, peerIP, ipv6=False, globalPeer=False, present=True):
        """Check that a route is present/not present in a (in-cluster) calico-node bird instance"""
        def fn():
            birdCmd = "birdcl6" if ipv6 else "birdcl"
            birdPeer = "Global_" if globalPeer else "Node_"
            birdPeer += peerIP.replace(".", "_").replace(":","_")
            routes = kubectl("exec -n calico-system %s -- %s show route protocol %s" % (calicoPod, birdCmd, birdPeer))
            result = re.search(r"%s *via %s on .* \[%s" % (re.escape(route), re.escape(peerIP), birdPeer), routes)
            if result is None and present:
                raise Exception('route not present when it should be')
            if result is not None and not present:
                raise Exception('route present when it should not be')
            return result
        result = retry_until_success(fn, wait_time=3)
        return result

    def _assert_route_present_in_cluster_bird(self, calicoPod, route, peerIP, ipv6=False, globalPeer=False):
        self._check_route_in_cluster_bird(calicoPod, route, peerIP, ipv6=ipv6, globalPeer=globalPeer, present=True)

    def _assert_route_not_present_in_cluster_bird(self, calicoPod, route, peerIP, ipv6=False, globalPeer=False):
        self._check_route_in_cluster_bird(calicoPod, route, peerIP, ipv6=ipv6, globalPeer=globalPeer, present=False)

    def _check_route_in_external_bird(self, birdContainer, birdPeer, routeRegex, peerIPRegex, ipv6=False, present=True):
        """Check that a route is present/not present in an external (plain docker container) bird instance"""
        def fn():
            birdCmd = "birdcl6" if ipv6 else "birdcl"
            routes = run("docker exec %s %s show route protocol %s" % (birdContainer, birdCmd, birdPeer))
            result = re.search(r"%s *via %s on .* \[%s" % (routeRegex, peerIPRegex, birdPeer), routes)
            if result is None and present:
                raise Exception('route not present when it should be')
            if result is not None and not present:
                raise Exception('route present when it should not be')
            return result
        result = retry_until_success(fn, wait_time=3)
        return result

    def _assert_route_present_in_external_bird(self, birdContainer, birdPeer, routeRegex, peerIPRegex, ipv6=False):
        self._check_route_in_external_bird(birdContainer, birdPeer, routeRegex, peerIPRegex, ipv6=ipv6, present=True)

    def _assert_route_not_present_in_external_bird(self, birdContainer, birdPeer, routeRegex, peerIPRegex, ipv6=False):
        self._check_route_in_external_bird(birdContainer, birdPeer, routeRegex, peerIPRegex, ipv6=ipv6, present=False)

    def _patch_peer_filters(self, peer, filters):
        """Patch BGPFilters in a BGPPeer"""
        filterStr = "\"" + "\", \"".join(filters) + "\"" if len(filters) > 0 else ""
        patchStr = "{\"spec\": {\"filters\": [%s]}}" % filterStr
        kubectl("patch --type merge bgppeer %s --patch '%s'" % (peer, patchStr))


    def _test_bgp_filter_basic(self, ipv4, ipv6):
        """Basic test case:
        - Add IPv4/IPv6 route to the external bird instance
        - Verify it is present in cluster bird instance (to validate import)
        - Verify that IPAM block from IP pool is present in external bird instance (to validate export)
        - Add BGPFilter with export reject rule and verify that route is no longer present in external bird instance
        - Add BGPFilter with import reject rule and verify that route is no longer present in cluster bird instance
        """
        with DiagsCollector():
            external_route_v4 = "10.111.111.0/24"
            cluster_route_regex_v4 = r"192\.168\.\d+\.\d+/\d+"
            export_filter_cidr_v4 = "192.168.0.0/16"

            external_route_v6 = "fd00:1111:1111:1111::/64"
            cluster_route_regex_v6 = r"fd00:10:244:.*/\d+"
            export_filter_cidr_v6 = "fd00:10:244::/64"

            # Add static route bird config to external node
            if ipv4:
                run("""cat <<EOF | docker exec -i kube-node-extra sh -c "cat > /etc/bird/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v4, self.external_node_ip))
                run("docker exec kube-node-extra birdcl configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra sh -c 'rm /etc/bird/static-route.conf; birdcl configure'"))
            if ipv6:
                run("""cat <<EOF | docker exec -i kube-node-extra-v6 sh -c "cat > /etc/bird6/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v6, self.external_node_ip6))
                run("docker exec kube-node-extra-v6 birdcl6 configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra-v6 sh -c 'rm /etc/bird6/static-route.conf; birdcl6 configure'"))

            # Check that the egress node (i.e., the cluster node that peers through BGP with the external bird instance) has the route advertised by the external bird instance
            if ipv4:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip)
            if ipv6:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, ipv6=True)

            # Check that the external bird instance has a route for an IPAM block from the default IP pool
            if ipv4:
                self._assert_route_present_in_external_bird("kube-node-extra", "Mesh_with_node_1", cluster_route_regex_v4, re.escape(self.egress_node_ip))
            if ipv6:
                self._assert_route_present_in_external_bird("kube-node-extra-v6", "Mesh_with_node_1", cluster_route_regex_v6, "2001:20::1", ipv6=True)
            # Add BGPFilter with export rule and check that the external bird instance no longer has the route for an IPAM block from the default IP pool
            if ipv4:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-export-1
spec:
  exportV4:
  - cidr: %s
    matchOperator: In
    action: Reject
EOF
""" % export_filter_cidr_v4)
                self._patch_peer_filters("node-extra.peer", ["test-filter-export-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-export-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))

                self._assert_route_not_present_in_external_bird("kube-node-extra", "Mesh_with_node_1", cluster_route_regex_v4, re.escape(self.egress_node_ip))
            if ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-export-v6-1
spec:
  exportV6:
  - cidr: %s
    matchOperator: In
    action: Reject
EOF
""" % export_filter_cidr_v6)
                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-export-v6-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-export-v6-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))

                self._assert_route_not_present_in_external_bird("kube-node-extra-v6", "Mesh_with_node_1", cluster_route_regex_v6, "2001:20::1", ipv6=True)

            # Add BGPFilter with import rule and check that the egress node no longer has the route advertised by the external bird instance
            if ipv4:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-1
spec:
  importV4:
  - cidr: %s
    matchOperator: Equal
    action: Reject
EOF
""" % external_route_v4)
                self._patch_peer_filters("node-extra.peer", ["test-filter-import-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))

                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip)
            if ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-v6-1
spec:
  importV6:
  - cidr: %s
    matchOperator: Equal
    action: Reject
EOF
""" % external_route_v6)
                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-import-v6-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-v6-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))

                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, ipv6=True)

    def _test_bgp_filter_ordering(self, ipv4, ipv6):
        """Test multiple rules per filter and multiple filters per peer, as well as
        exhaust matchOperators and actions"""
        with DiagsCollector():
            external_route_v4 = "10.111.111.0/24"
            export_filter_cidr_v4 = "192.168.0.0/16"

            external_route_v6 = "fd00:1111:1111:1111::/64"
            export_filter_cidr_v6 = "fd00:10:244::/64"

            # Add static route bird config
            if ipv4:
                run("""cat <<EOF | docker exec -i kube-node-extra sh -c "cat > /etc/bird/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v4, self.external_node_ip))
                run("docker exec kube-node-extra birdcl configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra sh -c 'rm /etc/bird/static-route.conf; birdcl configure'"))
            if ipv6:
                run("""cat <<EOF | docker exec -i kube-node-extra-v6 sh -c "cat > /etc/bird6/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v6, self.external_node_ip6))
                run("docker exec kube-node-extra-v6 birdcl6 configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra-v6 sh -c 'rm /etc/bird6/static-route.conf; birdcl6 configure'"))

            # Add filters with multiple rules that should result in the routes being accepted
            if ipv4:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-1
spec:
  importV4:
  - cidr: 10.111.0.0/16
    matchOperator: In
    action: Accept
  - cidr: 10.111.111.0/24
    matchOperator: Equal
    action: Reject
EOF
""")
                self._patch_peer_filters("node-extra.peer", ["test-filter-import-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))
                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-1"))
            if ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-v6-1
spec:
  importV6:
  - cidr: fd00:1111:1111::/48
    matchOperator: In
    action: Accept
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: Equal
    action: Reject
EOF
""")
                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-import-v6-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))
                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-v6-1"))

            # Check that routes are present
            if ipv4:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip)
            if ipv6:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, ipv6=True)

            # Add additional filters with multiple rules that should result in the routes being rejected
            if ipv4:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-2
spec:
  importV4:
  - cidr: 10.111.0.0/16
    matchOperator: NotIn
    action: Accept
  - cidr: 10.111.111.0/24
    matchOperator: NotEqual
    action: Accept
  - cidr: 10.111.111.0/24
    matchOperator: Equal
    action: Reject
EOF
""")
                self._patch_peer_filters("node-extra.peer", ["test-filter-import-2", "test-filter-import-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))
                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-2"))
            if ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-v6-2
spec:
  importV6:
  - cidr: fd00:1111:1111::/48
    matchOperator: NotIn
    action: Accept
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: NotEqual
    action: Accept
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: Equal
    action: Reject
EOF
""")
                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-import-v6-2", "test-filter-import-v6-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))
                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-v6-2"))

            # Check that routes are no longer present
            if ipv4:
                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip)
            if ipv6:
                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, ipv6=True)

            # Add an additional filter with both IPv4 and IPv6 rules that should result in the routes being accepted
            if ipv4 and ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-v4-v6
spec:
  importV4:
  - cidr: 10.111.111.0/24
    matchOperator: Equal
    action: Accept
  - cidr: 10.111.0.0/16
    matchOperator: NotIn
    action: Accept
  - cidr: 10.111.111.0/24
    matchOperator: NotEqual
    action: Accept
  - cidr: 10.111.111.0/24
    matchOperator: Equal
    action: Reject
  importV6:
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: Equal
    action: Accept
  - cidr: fd00:1111:1111::/48
    matchOperator: NotIn
    action: Accept
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: NotEqual
    action: Accept
  - cidr: fd00:1111:1111:1111::/64
    matchOperator: Equal
    action: Reject
EOF
""")
                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-v4-v6"))

                self._patch_peer_filters("node-extra.peer", ["test-filter-import-v4-v6", "test-filter-import-2", "test-filter-import-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))

                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-import-v4-v6", "test-filter-import-v6-2", "test-filter-import-v6-1"])
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))

                # Check that routes are present
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip)
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, ipv6=True)

    def _test_bgp_filter_global_peer(self, ipv4, ipv6):
        """Test BGP import filters with global BGP peers"""
        with DiagsCollector():
            external_route_v4 = "10.111.111.0/24"
            export_filter_cidr_v4 = "192.168.0.0/16"

            external_route_v6 = "fd00:1111:1111:1111::/64"
            export_filter_cidr_v6 = "fd00:10:244::/64"

            # Add static route bird config
            if ipv4:
                run("""cat <<EOF | docker exec -i kube-node-extra sh -c "cat > /etc/bird/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v4, self.external_node_ip))
                run("docker exec kube-node-extra birdcl configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra sh -c 'rm /etc/bird/static-route.conf; birdcl configure'"))
            if ipv6:
                run("""cat <<EOF | docker exec -i kube-node-extra-v6 sh -c "cat > /etc/bird6/static-route.conf"
protocol static static1 {
    route %s via %s;
    export all;
}
EOF
""" % (external_route_v6, self.external_node_ip6))
                run("docker exec kube-node-extra-v6 birdcl6 configure")
                self.add_cleanup(lambda: run("docker exec kube-node-extra-v6 sh -c 'rm /etc/bird6/static-route.conf; birdcl6 configure'"))

            # Patch BGPPeer to make it global
            if ipv4:
                kubectl("patch --type=merge bgppeer node-extra.peer --type json --patch '[{\"op\": \"remove\", \"path\": \"/spec/nodeSelector\"}]'")
                self.add_cleanup(lambda: kubectl("patch --type=merge bgppeer node-extra.peer --patch '{\"spec\":{\"nodeSelector\":\"egress == \\\"true\\\"\"}}'"))
            if ipv6:
                kubectl("patch --type=merge bgppeer node-extra-v6.peer --type json --patch '[{\"op\": \"remove\", \"path\": \"/spec/nodeSelector\"}]'")
                self.add_cleanup(lambda: kubectl("patch --type=merge bgppeer node-extra-v6.peer --patch '{\"spec\":{\"nodeSelector\":\"egress == \\\"true\\\"\"}}'"))

            # Check that route is present
            if ipv4:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip, globalPeer=True)
            if ipv6:
                self._assert_route_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, globalPeer=True, ipv6=True)

            # Add BGPFilter with import rule and check that the route is no longer present
            if ipv4:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-1
spec:
  importV4:
  - cidr: %s
    matchOperator: Equal
    action: Reject
EOF
""" % external_route_v4)
                self._patch_peer_filters("node-extra.peer", ["test-filter-import-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra.peer", []))

                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v4, self.external_node_ip, globalPeer=True)
            if ipv6:
                kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-import-v6-1
spec:
  importV6:
  - cidr: %s
    matchOperator: Equal
    action: Reject
EOF
""" % external_route_v6)
                self._patch_peer_filters("node-extra-v6.peer", ["test-filter-import-v6-1"])

                self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-import-v6-1"))
                self.add_cleanup(lambda: self._patch_peer_filters("node-extra-v6.peer", []))

                self._assert_route_not_present_in_cluster_bird(self.egress_calico_pod, external_route_v6, self.external_node_ip6, globalPeer=True, ipv6=True)

    def test_bgp_filter_basic_v4(self):
        self._test_bgp_filter_basic(True, False)

    def test_bgp_filter_basic_v6(self):
        self._test_bgp_filter_basic(False, True)

    def test_bgp_filter_basic_v4v6(self):
        self._test_bgp_filter_basic(True, True)

    def test_bgp_filter_ordering_v4(self):
        self._test_bgp_filter_ordering(True, False)

    def test_bgp_filter_ordering_v6(self):
        self._test_bgp_filter_ordering(False, True)

    def test_bgp_filter_ordering_v4v6(self):
        self._test_bgp_filter_ordering(True, True)

    def test_bgp_filter_global_peer_v4(self):
        self._test_bgp_filter_global_peer(True, False)

    def test_bgp_filter_global_peer_v6(self):
        self._test_bgp_filter_global_peer(False, True)

    def test_bgp_filter_global_peer_v4v6(self):
        self._test_bgp_filter_global_peer(True, True)



# Copyright 2019 Tigera, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from etcd3gw.exceptions import Etcd3Exception

import mock

from neutron.tests import base

from networking_calico.agent.dhcp_agent import SubnetWatcher
from networking_calico.etcdutils import EtcdWatcher


LOG = logging.getLogger(__name__)


class TestSubnetWatcher(base.BaseTestCase):

    @mock.patch.object(EtcdWatcher, "start")
    def test_exception_detail_logging(self, loop_fn):

        # Make EtcdWatcher.start throw an exception with detail text.
        loop_fn.side_effect = Etcd3Exception(detail_text="from test_exception_detail")

        with mock.patch.object(
            logging.getLogger("networking_calico.agent.dhcp_agent"), "exception"
        ) as mock_le:
            # Create the DHCP agent and allow it to start the
            # SubnetWatcher loop.
            sw = SubnetWatcher(mock.Mock(), "/calico")
            try:
                sw.start()
            except Exception:
                pass
            mock_le.assert_called_with(
                "Etcd3Exception in SubnetWatcher.start():\n%s",
                "from test_exception_detail",
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_config import cfg

from networking_calico import datamodel_v2
from networking_calico import datamodel_v3


DEFAULT_BW_BURST = 4294967296
DEFAULT_PR_BURST = 5

SHARED_OPTS = [
    # etcd connection information.
    cfg.StrOpt(
        "etcd_host",
        default="127.0.0.1",
        help="The hostname or IP of the etcd node/proxy",
    ),
    cfg.IntOpt(
        "etcd_port", default=2379, help="The port to use for the etcd node/proxy"
    ),
    cfg.StrOpt(
        "etcd_scheme",
        default="http",
        help="The protocol scheme to be used for connections to etcd",
    ),
    # etcd TLS-related options.
    cfg.StrOpt("etcd_key_file", help="The path to the TLS key file to use with etcd."),
    cfg.StrOpt(
        "etcd_cert_file",
        help="The path to the TLS client certificate file to use with etcd.",
    ),
    cfg.StrOpt(
        "etcd_ca_cert_file",
        help="The path to the TLS CA certificate file to use with etcd.",
    ),
    cfg.StrOpt(
        "etcd_username",
        help="User name for accessing an etcd cluster with authentication enabled.",
    ),
    cfg.StrOpt(
        "etcd_password",
        help="Password for accessing an etcd cluster with authentication enabled.",
    ),
    # Large etcd subtree snapshot reads can take time, hence the
    # default of 60 seconds here as opposed to something much shorter.
    cfg.IntOpt(
        "etcd_timeout", default=60, help="Timeout (in seconds) for etcd requests."
    ),
    cfg.StrOpt(
        "openstack_region",
        help=(
            "When in a multi-region OpenStack deployment, a unique "
            "name for the region that this node (controller or "
            "compute) belongs to."
        ),
    ),
    # Options for QoS parameters that are supported on the Calico
    # WorkloadEndpoint resource but not (yet) represented on the Neutron API.
    #
    # The complete mapping between OpenStack-level config/API and the Calico
    # WorkloadEndpoint.QoSControls is as follows.
    #
    #  QoSControls field     | Neutron API field     | Config field
    # -----------------------+-----------------------+----------------------------------
    #  IngressBandwidth      | max_kbps * 1000       |
    #  EgressBandwidth       | max_kbps * 1000       |
    #  IngressBurst          |                       | ingress_burst_bits
    #  EgressBurst           |                       | egress_burst_bits
    #  IngressPeakrate       | max_burst_kbps * 1000 |
    #  EgressPeakrate        | max_burst_kbps * 1000 |
    #  IngressMinburst       |                       | ingress_minburst_bytes
    #  EgressMinburst        |                       | egress_minburst_bytes
    #  IngressPacketRate     | max_kpps * 1000       |
    #  EgressPacketRate      | max_kpps * 1000       |
    #  IngressPacketBurst    |                       | ingress_burst_packets
    #  EgressPacketBurst     |                       | egress_burst_packets
    #   (not implemented)    | max_burst_kpps        |
    #  IngressMaxConnections |                       | max_ingress_connections_per_port
    #  EgressMaxConnections  |                       | max_egress_connections_per_port
    #
    # Note, max_burst_kpps is not currently implemented, because we have not
    # yet found a reasonable way to do that.
    cfg.IntOpt(
        "max_ingress_connections_per_port",
        default=0,
        help=(
            "If non-zero, a maximum number of ingress connections to impose on each"
            " port."
        ),
    ),
    cfg.IntOpt(
        "max_egress_connections_per_port",
        default=0,
        help=(
            "If non-zero, a maximum number of egress connections to impose on each"
            " port."
        ),
    ),
    cfg.IntOpt(
        "ingress_burst_bits",
        default=DEFAULT_BW_BURST,
        help=(
            "If non-zero, configures the maximum allowed burst at peakrate, in the"
            " ingress direction."
        ),
    ),
    cfg.IntOpt(
        "egress_burst_bits",
        default=DEFAULT_BW_BURST,
        help=(
            "If non-zero, configures the maximum allowed burst at peakrate, in the"
            " egress direction."
        ),
    ),
    cfg.IntOpt(
        "ingress_minburst_bytes",
        default=0,
        help=(
            "If non-zero, configures the minimum burst size for peakrate data, in the"
            " ingress direction."
        ),
    ),
    cfg.IntOpt(
        "egress_minburst_bytes",
        default=0,
        help=(
            "If non-zero, configures the minimum burst size for peakrate data, in the"
            " egress direction."
        ),
    ),
    cfg.IntOpt(
        "ingress_burst_packets",
        default=DEFAULT_PR_BURST,
        help=(
            "If non-zero, configures the maximum allowed packet rule burst, in the"
            " ingress direction."
        ),
    ),
    cfg.IntOpt(
        "egress_burst_packets",
        default=DEFAULT_PR_BURST,
        help=(
            "If non-zero, configures the maximum allowed packet rule burst, in the"
            " egress direction."
        ),
    ),
]


def register_options(conf, additional_options=None):
    options_to_register = (
        SHARED_OPTS if additional_options is None else SHARED_OPTS + additional_options
    )
    conf.register_opts(options_to_register, "calico")


_cached_region_string = None
MAX_DNS_LABEL_LEN = 63
MAX_REGION_LEN = MAX_DNS_LABEL_LEN - len(
    datamodel_v3.REGION_NAMESPACE_PREFIX + datamodel_v2.REGION_PREFIX
)
DNS_LABEL_REGEXP = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?$"


def _validate_region(region):
    assert len(region) <= MAX_REGION_LEN, (
        "The value of openstack_region must be must be %d chars or fewer"
        % MAX_REGION_LEN
    )
    assert re.compile(DNS_LABEL_REGEXP).match(region), (
        "The value of openstack_region must be must be a valid DNS label;"
        " comprising lower case alphanumeric characters or '-',"
        " and starting and ending with an alphanumeric character"
    )


def get_region_string():
    """Return a per-region string for insertion into etcd key paths.

    Some etcd key paths that are only used with OpenStack need to be
    made unique per-region.  Previously these were:

    - /calico/felix/v1/..., for reporting Felix agent and endpoint
      status

    - /calico/openstack/v1/..., for electing a leader among the
      possibly multiple Neutron driver instances within a region

    - /calico/dhcp/v1/..., for passing Neutron subnet information from
      the Neutron driver to the Calico DHCP agent.

    With the introduction of multi-region support, these become:

    - /calico/felix/v2/<region_string>/...

    - /calico/openstack/v2/<region_string>/...

    - /calico/dhcp/v2/<region_string>/...

    where <region_string> is as returned by this function.
    """
    global _cached_region_string
    if _cached_region_string is None:
        # Use [calico] openstack_region if configured.
        if cfg.CONF.calico.openstack_region:
            _validate_region(cfg.CONF.calico.openstack_region)
            _cached_region_string = "%s%s" % (
                datamodel_v2.REGION_PREFIX,
                cfg.CONF.calico.openstack_region,
            )
        else:
            _cached_region_string = datamodel_v2.NO_REGION
    return _cached_region_string


def _reset_globals():
    global _cached_region_string
    _cached_region_string = None



# Copyright (c) 2015-2016 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import os
import re
import socket
import sys
import tempfile
from datetime import datetime
from subprocess import CalledProcessError
from subprocess import check_output, STDOUT

import json
import logging
from pprint import pformat

import yaml
from deepdiff import DeepDiff

LOCAL_IP_ENV = "MY_IP"
LOCAL_IPv6_ENV = "MY_IPv6"
logger = logging.getLogger(__name__)

ETCD_SCHEME = os.environ.get("ETCD_SCHEME", "http")
ETCD_CA = os.environ.get("ETCD_CA_CERT_FILE", "")
ETCD_CERT = os.environ.get("ETCD_CERT_FILE", "")
ETCD_KEY = os.environ.get("ETCD_KEY_FILE", "")
ETCD_HOSTNAME_SSL = "etcd-authority-ssl"
KUBECONFIG = "/home/user/certs/kubeconfig"

API_VERSION = 'projectcalico.org/v3'
ERROR_CONFLICT = "update conflict"
NOT_FOUND = "resource does not exist"
NOT_NAMESPACED = "is not namespaced"
SET_DEFAULT = "Cannot set"
NOT_SUPPORTED = "is not supported on"
KUBERNETES_NP = "kubernetes network policies must be managed through the kubernetes API"
NOT_LOCKED = "datastore is not locked. Run the `calicoctl datastore migrate lock` command in order to begin migration"
NOT_KUBERNETES = "invalid datastore type: etcdv3 to import to for datastore migration. Datastore type must be kubernetes"
NO_IPAM = "no IPAM resources specified in file"
NOT_LOCKED_SPLIT = "datastore is not locked. Run the `calicoctl datastore migrate lock` command in order split the IP pools"
POOL_NOT_EXIST_CIDR = "unable to find IP pool"
INVALID_SPLIT_NUM = "number to split CIDR into is not a valid power of 2"
POOL_TOO_SMALL = "is not large enough to be split into"


class CalicoctlOutput:
    """
    CalicoctlOutput contains the output from running a calicoctl command using
    the calicoctl function below.

    This class contains the command, output and error code (if it failed)
    along with YAML/JSON decoded output if the output could be decoded.
    """

    def __init__(self, command, output, error=None):
        self.command = command
        self.output = output
        self.error = error

        # Attempt to decode the output and store the output format.
        self.decoded, self.decoded_format = decode_json_yaml(self.output)

    def assert_data(self, data, format="yaml", text=None):
        """
        Assert the decoded output from the calicoctl command matches the
        supplied data and the expected decoder format.
        Args:
            data:   The data to compare
            format: The expected output format of the data.
            text:   (optional) Expected text in the command output.
        """
        self.assert_no_error(text)
        assert self.decoded is not None, "No value was decoded from calicoctl response."
        if isinstance(data, str):
            data, _ = decode_json_yaml(data)
            assert data is not None, "String data did not decode"

        if format is not None:
            assert format == self.decoded_format, "Decoded format is different. " \
                                                  "expect %s; got %s" % (format, self.decoded_format)

        # Copy and clean the decoded data to allow it to be comparable.
        cleaned = clean_calico_data(self.decoded)

        print(self.decoded)
        assert cleaned == data, \
            "Items are not the same.  Difference is:\n %s" % \
            pformat(DeepDiff(cleaned, data), indent=2)

    def assert_empty_list(self, kind, format="yaml", text=None):
        """
        Assert the calicoctl command output an empty list of the specified
        kind.

        Args:
            kind:   The resource kind.
            format: The expected output format of the data.
            text:   (optional) Expected text in the command output.

        Returns:

        """
        data = make_list(kind, [])
        self.assert_data(data, format=format, text=text)

    def assert_list(self, kind, items, format="yaml", text=None):
        """
        Assert the calicoctl command output a list of the specified
        kind.

        Args:
            kind:   The resource kind.
            items:  A list of the items in the list.
            format: The expected output format of the data.
            text:   (optional) Expected text in the command output.

        Returns:

        """
        data = make_list(kind, items)
        self.assert_data(data, format=format, text=text)

    def assert_error(self, text=None):
        """
        Assert the calicoctl command exited with an error and did not panic
        Args:
            text:   (optional) Expected text in the comma, another_resource)
        rc.assert_no_error()

        rc = calicoctl("get %s -nd output.
        """
        assert self.error, "Expected error running command; \n" \
                           "command=" + self.command + "\noutput=" + self.output
        assert not "panic" in self.output, "Exited with an error due to a panic"
        self.assert_output_contains(text)

    def assert_no_error(self, text=None):
        """
        Assert the calicoctl command did not exit with an error code.
        Args:
            text:   (optional) Expected text in the command output.
        """
        assert not self.error, "Expected no error running command; \n" \
                               "command=" + self.command + "\noutput=" + self.output

        # If text is supplied, assert it appears in the output
        if text:
            self.assert_output_contains(text)

    def assert_output_equals(self, text):
        """
        Assert the calicoctl command output is exactly the supplied text.
        Args:
            text:   Expected text in the command output.
        """
        if not text:
            return
        assert text == self.output, "Expected output to exactly match; \n" + \
                                    "command=" + self.command + "\noutput=\n" + self.output + \
                                    "\nexpected=\n" + text

    def assert_output_equals_ignore_res_version(self, text):
        """
        Assert the calicoctl command output is exactly the supplied text.
        Args:
            text:   Expected text in the command output.
        """
        if not text:
            return

        text = re.sub('resourceVersion: ".*?"', 'resourceVersion: "<ignored>"', text)
        out = re.sub('resourceVersion: ".*?"', 'resourceVersion: "<ignored>"', self.output)

        assert text == out, "Expected output to match after ignoring resource version; \n" + \
                                    "command=" + self.command + "\noutput=\n" + out + \
                                    "\nexpected=\n" + text

    def assert_output_contains(self, text):
        """
        Assert the calicoctl command output contains the supplied text.
        Args:
            text:   Expected text in the command output.
        """
        if not text:
            return
        assert text in self.output, "Expected text in output; \n" + \
                                    "command=" + self.command + "\noutput=\n" + self.output + \
                                    "\nexpected=\n" + text

    def assert_output_not_contains(self, text):
        """
        Assert the calicoctl command output does not contain the supplied text.
        Args:
            text:   Expected text in the command output.
        """
        if not text:
            return
        assert not text in self.output, "Unexpected text in output; \n" + \
                                        "command=" + self.command + "\noutput=\n" + self.output + \
                                        "\nunexpected=\n" + text


def calicoctl(command, data=None, load_as_stdin=False, format="yaml", only_stdout=False, no_config=False, kdd=False, allowVersionMismatch=True):
    """
    Convenience function for abstracting away calling the calicoctl
    command.

    :param command:  The calicoctl command line parms as a single string.
    :param data:  Input data either as a string or a JSON serializable Python
    object.
    :param load_as_stdin:  Load the input data through stdin rather than by
    loading from file.
    :param format:  Specify the format for loading the data.
    :param only_stdout: Return only the stdout
    :return: The output from the command with leading and trailing
    whitespace removed.
    """
    # If input data is specified, save it to file in the required format.
    if isinstance(data, str):
        data, _ = decode_json_yaml(data)
        assert data is not None, "String data did not decode"
    if data is not None:
        if format == "yaml":
            writeyaml("/tmp/input-data", data)
        else:
            writejson("/tmp/input-data", data)

    stdin = ''
    option_file = ''

    if data and load_as_stdin:
        stdin = 'cat /tmp/input-data | '
        option_file = ' -f -'
    elif data and not load_as_stdin:
        option_file = ' -f /tmp/input-data'

    calicoctl_bin = os.environ.get("CALICOCTL", "/code/bin/calicoctl-linux-amd64")

    if allowVersionMismatch:
        calicoctl_bin += " --allow-version-mismatch"

    if ETCD_SCHEME == "https":
        etcd_auth = "%s:2379" % ETCD_HOSTNAME_SSL
    else:
        etcd_auth = "%s:2379" % get_ip()

    # Export the environment, in case the command has multiple parts, e.g.
    # use of | or ;
    #
    # Pass in all etcd params, the values will be empty if not set anyway
    calicoctl_env_cmd = "export ETCD_ENDPOINTS=%s; " \
                        "export ETCD_CA_CERT_FILE=%s; " \
                        "export ETCD_CERT_FILE=%s; " \
                        "export ETCD_KEY_FILE=%s; " \
                        "export DATASTORE_TYPE=%s; %s %s" % \
                        (ETCD_SCHEME + "://" + etcd_auth, ETCD_CA, ETCD_CERT, ETCD_KEY,
                         "etcdv3", stdin, calicoctl_bin)
    if kdd:
        calicoctl_env_cmd = "export DATASTORE_TYPE=kubernetes; " \
                            "export KUBECONFIG=%s; %s %s" % \
                            (KUBECONFIG, stdin, calicoctl_bin)
    if no_config:
        calicoctl_env_cmd = stdin + calicoctl_bin
    full_cmd = calicoctl_env_cmd + " " + command + option_file

    try:
        output = log_and_run(full_cmd, stderr=(None if only_stdout else STDOUT))
        return CalicoctlOutput(full_cmd, output)
    except CalledProcessError as e:
        return CalicoctlOutput(full_cmd, e.output.decode(), error=e.returncode)


def clean_calico_data(data, extra_keys_to_remove=None):
    """
    Clean the data returned from a calicoctl get command to remove empty
    structs, null values and non-configurable fields.  This makes comparison
    with the input data much simpler.

    Args:
        data: The data to clean.
        extra_keys_to_remove: more keys to remove if needed.

    Returns: The cleaned data.

    """
    new = copy.deepcopy(data)

    # Recursively delete empty structs / nil values and non-configurable
    # fields.
    def clean_elem(elem, extra_keys):
        if isinstance(elem, list):
            # Loop through each element in the list
            for i in elem:
                clean_elem(i, extra_keys)
        if isinstance(elem, dict):
            # Remove non-settable fields, and recursively clean each value of
            # the dictionary, removing nil values or values that are empty
            # dicts after cleaning.
            del_keys = ['creationTimestamp', 'resourceVersion', 'uid']
            if extra_keys is not None:
                for extra_key in extra_keys:
                    del_keys.append(extra_key)
            for k, v in elem.items():
                clean_elem(v, extra_keys)
                if v is None or v == {}:
                    del_keys.append(k)
            for k in del_keys:
                if k in elem:
                    del (elem[k])

    clean_elem(new, extra_keys_to_remove)
    return new


def add_tier_label(data):
    """
    Convenience method for auto-adding the `projectcalico.org/tier`.
    """
    new = copy.deepcopy(data)

    def add_label(elem):
        if isinstance(elem, list):
            for i in elem:
                add_label(i)
        if isinstance(elem, dict):
            if elem['kind'] not in ['NetworkPolicy', 'GlobalNetworkPolicy', 'StagedNetworkPolicy',
                                    'StagedGlobalNetworkPolicy']:
                return
            tier = 'default'
            if 'tier' in elem['spec']:
                tier = elem['spec']['tier']
            if 'labels' not in elem['metadata']:
                elem['metadata']['labels'] = {}
            elem['metadata']['labels']['projectcalico.org/tier'] = tier

    add_label(new)
    return new


def decode_json_yaml(value):
    try:
        decoded = json.loads(value)
        # fix the python datetime back into isoformat with empty timezone information
        decoded = find_and_format_creation_timestamp(decoded)
        return decoded, "json"
    except ValueError:
        pass
    try:
        decoded = yaml.safe_load(value)
        # fix the python datetime back into isoformat with empty timezone information
        decoded = find_and_format_creation_timestamp(decoded)
        return decoded, "yaml"
    except yaml.YAMLError:
        pass
    return None, None


def find_and_format_creation_timestamp(decoded):
    if decoded:
        if 'items' in decoded:
            for i in range(len(decoded['items'])):
                decoded['items'][i] = format_creation_timestamp(decoded['items'][i])
        else:
            decoded = format_creation_timestamp(decoded)
    return decoded


def format_creation_timestamp(decoded):
    if isinstance(decoded, dict) and 'metadata' in decoded and 'creationTimestamp' in decoded['metadata']:
        if isinstance(decoded['metadata']['creationTimestamp'], datetime):
            decoded['metadata']['creationTimestamp'] = decoded.get('metadata', {}). \
                                                           get('creationTimestamp', datetime.utcnow()).isoformat() + 'Z'
    return decoded


def writeyaml(filename, data):
    """
    Converts a python dict to yaml and outputs to a file.
    :param filename: filename to write
    :param data: dictionary to write out as yaml
    """
    with open(filename, 'w') as f:
        text = yaml.dump(data, default_flow_style=False)
        logger.debug("Writing %s: \n%s" % (filename, truncate_for_log(text, 4000)))
        f.write(text)


def writejson(filename, data):
    """
    Converts a python dict to json and outputs to a file.
    :param filename: filename to write
    :param data: dictionary to write out as json
    """
    with open(filename, 'w') as f:
        text = json.dumps(data,
                          sort_keys=True,
                          indent=2,
                          separators=(',', ': '))
        logger.debug("Writing %s: \n%s" % (filename, truncate_for_log(text, 4000)))
        f.write(text)


def truncate_for_log(text, length):
    if len(text) <= length:
        return text
    return text[:length] + "... <truncated>"


def get_ip(v6=False):
    """
    Return a string of the IP of the hosts interface.
    Try to get the local IP from the environment variables.  This allows
    testers to specify the IP address in cases where there is more than one
    configured IP address for the test system.
    """
    env = LOCAL_IPv6_ENV if v6 else LOCAL_IP_ENV
    ip = os.environ.get(env)
    if not ip:
        logger.debug("%s not set; try to auto detect IP.", env)
        socket_type = socket.AF_INET6 if v6 else socket.AF_INET
        s = socket.socket(socket_type, socket.SOCK_DGRAM)
        remote_ip = "2001:4860:4860::8888" if v6 else "8.8.8.8"
        s.connect((remote_ip, 0))
        ip = s.getsockname()[0]
        s.close()
    else:
        logger.debug("Got local IP from %s=%s", env, ip)

    return ip


def log_and_run(command, raise_exception_on_failure=True, stderr=STDOUT):
    def log_output(results):
        if results is None:
            logger.info("  # <no output>")

        lines = results.split("\n")
        for line in lines:
            logger.info("  # %s", line.rstrip())

    try:
        logger.info("%s", command)
        results = check_output(command, shell=True, stderr=stderr).rstrip()
        results = results.decode()
        log_output(results)
        return results
    except CalledProcessError as e:
        # Wrap the original exception with one that gives a better error
        # message (including command output).
        logger.info("  # Return code: %s", e.returncode)
        log_output(e.output.decode())
        if raise_exception_on_failure:
            raise e


def curl_etcd(path, options=None, recursive=True, ip=None):
    """
    Perform a curl to etcd, returning JSON decoded response.
    :param path:  The key path to query
    :param options:  Additional options to include in the curl
    :param recursive:  Whether we want recursive query or not
    :return:  The JSON decoded response.
    """
    if options is None:
        options = []
    if ETCD_SCHEME == "https":
        # Etcd is running with SSL/TLS, require key/certificates
        rc = check_output(
            "curl --cacert %s --cert %s --key %s "
            "-sL https://%s:2379/v2/keys/%s?recursive=%s %s"
            % (ETCD_CA, ETCD_CERT, ETCD_KEY, ETCD_HOSTNAME_SSL,
               path, str(recursive).lower(), " ".join(options)),
            shell=True)
    else:
        rc = check_output(
            "curl -sL http://%s:2379/v2/keys/%s?recursive=%s %s"
            % (ip, path, str(recursive).lower(), " ".join(options)),
            shell=True)

    logger.info("etcd RC: %s" % rc.strip())
    return json.loads(rc.strip())


def wipe_etcd(ip):
    # Delete /calico if it exists. This ensures each test has an empty data
    # store at start of day.
    curl_etcd("calico", options=["-XDELETE"], ip=ip)

    # Disable Usage Reporting to usage.projectcalico.org
    # We want to avoid polluting analytics data with unit test noise
    curl_etcd("calico/v1/config/UsageReportingEnabled",
              options=["-XPUT -d value=False"], ip=ip)

    etcd_container_name = "calico-etcd"
    tls_vars = ""
    if ETCD_SCHEME == "https":
        # Etcd is running with SSL/TLS, require key/certificates
        etcd_container_name = "calico-etcd-ssl"
        tls_vars = ("-e ETCDCTL_CACERT=/etc/calico/certs/ca.pem " +
                    "-e ETCDCTL_CERT=/etc/calico/certs/client.pem " +
                    "-e ETCDCTL_KEY=/etc/calico/certs/client-key.pem ")

    check_output("docker exec " + tls_vars + etcd_container_name + " etcdctl del --prefix /calico", shell=True)


def make_list(kind, items):
    """
    Convert the list of resources into a single List resource type.
    Args:
        items: A list of the resources in the List object.

    Returns:
        None
    """
    assert isinstance(items, list)
    if "List" not in kind:
        kind = kind + "List"
    return {
        'kind': kind,
        'apiVersion': API_VERSION,
        'items': items,
    }


def name(data):
    """
    Returns the name of the resource in the supplied data
    Args:
        data: A dictionary containing the resource.

    Returns: The resource name.
    """
    return data['metadata']['name']

def kind(data):
    """
    Returns the kind of the resource in the supplied data
    Args:
        data: A dictionary containing the resource.

    Returns: The resource kind.
    """
    return data['kind']

def namespace(data):
    """
    Returns the namespace of the resource in the supplied data
    Args:
       data: A dictionary containing the resource.

    Returns: The resource namespace.
    """
    return data['metadata']['namespace']

def set_cluster_version(calico_version="", kdd=False):
    """
    Set Calico version in ClusterInformation using the calico_version_helper go app.
    Args:
        calico_version: string with version to set
        kdd: optional bool to indicate use of kubernetes datastore (default False)

    Returns: The command output
    """

    if ETCD_SCHEME == "https":
        etcd_auth = "%s:2379" % ETCD_HOSTNAME_SSL
    else:
        etcd_auth = "%s:2379" % get_ip()

    calico_helper_bin = "/code/tests/fv/helper/bin/calico_version_helper"
    full_cmd = "export ETCD_ENDPOINTS=%s; " \
        "export ETCD_CA_CERT_FILE=%s; " \
        "export ETCD_CERT_FILE=%s; " \
        "export ETCD_KEY_FILE=%s; " \
        "export DATASTORE_TYPE=%s; %s" % \
        (ETCD_SCHEME+"://"+etcd_auth, ETCD_CA, ETCD_CERT, ETCD_KEY,
         "etcdv3", calico_helper_bin)
    if kdd:
        full_cmd = "export DATASTORE_TYPE=kubernetes; " \
            "export KUBECONFIG=%s; %s" % \
            (KUBECONFIG, calico_helper_bin)
    if calico_version:
        full_cmd += " -v " + calico_version

    try:
        output = log_and_run(full_cmd, stderr=STDOUT)
        return CalicoctlOutput(full_cmd, output)
    except CalledProcessError as e:
        return CalicoctlOutput(full_cmd, e.output.decode(), error=e.returncode)



# -*- coding: utf-8 -*-
# Copyright (c) 2025 Tigera, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for CalicoMechanismDriver initialization and voting behavior.

These tests validate:
  - Only voting=True creates an Elector.
  - Voting=False never creates an Elector.
  - @requires_state does NOT cause a worker to become a voter.
  - A worker initialization does not override the parent elector.
"""

import unittest
import mock

import networking_calico.plugins.ml2.drivers.calico.test.lib as lib
from networking_calico.plugins.ml2.drivers.calico import mech_calico
from networking_calico import etcdv3


class TestMechanismDriverVoting(lib.Lib, unittest.TestCase):

    def setUp(self):
        super(TestMechanismDriverVoting, self).setUp()

        lib.m_oslo_config.cfg.CONF.keystone_authtoken.auth_url = ""
        lib.m_oslo_config.cfg.CONF.calico.openstack_region = "no-region"
        lib.m_oslo_config.cfg.CONF.calico.etcd_compaction_period_mins = 0
        lib.m_oslo_config.cfg.CONF.calico.resync_interval_secs = 0
        lib.m_oslo_config.cfg.CONF.calico.project_name_cache_max = 0

        # Mock etcd3gw client so background threads don't touch real etcd.
        etcdv3._client = self.clientv3 = mock.Mock()

        self.clientv3.status.return_value = {
            "header": {"revision": "123", "cluster_id": "cluster-id"},
        }
        self.clientv3.get_prefix.return_value = []
        self.clientv3.watch_prefix.return_value = (iter(()), mock.Mock())

        # Reset the driver
        mech_calico.mech_driver = None

    def tearDown(self):
        # Reset global etcd client.
        etcdv3._client = None
        mech_calico.mech_driver = None

        super(TestMechanismDriverVoting, self).tearDown()

    def _disable_background_threads(self, driver):
        """Disable background threads that would touch etcd or do unrelated things."""
        driver.periodic_resync_thread = mock.Mock()
        driver._status_updating_thread = mock.Mock()
        driver.resync_monitor_thread = mock.Mock()

    @mock.patch.object(mech_calico, "Elector")
    def test_parent_creates_elector(self, m_elector):
        driver = mech_calico.CalicoMechanismDriver()
        self._disable_background_threads(driver)

        driver._my_pid = None
        driver._post_fork_init(voting=True)

        m_elector.assert_called_once()
        self.assertIs(driver.elector, m_elector.return_value)

    @mock.patch.object(mech_calico, "Elector")
    def test_worker_does_not_create_elector(self, m_elector):
        driver = mech_calico.CalicoMechanismDriver()
        self._disable_background_threads(driver)

        driver._my_pid = None
        driver._post_fork_init(voting=False)

        m_elector.assert_not_called()
        self.assertIsNone(driver.elector)

    @mock.patch.object(mech_calico, "Elector")
    def test_requires_state_does_not_make_worker_voter(self, m_elector):
        driver = mech_calico.CalicoMechanismDriver()
        self._disable_background_threads(driver)

        driver._my_pid = None

        fake_context = mock.Mock()
        fake_context.original = {}
        fake_context.current = {}
        fake_context._plugin_context = mock.Mock()

        # update_network_postcommit is decorated with @requires_state
        driver.update_network_postcommit(fake_context)

        m_elector.assert_not_called()
        self.assertIsNone(driver.elector)

    @mock.patch.object(mech_calico, "Elector")
    def test_worker_init_does_not_override_parent_elector(self, m_elector):
        driver = mech_calico.CalicoMechanismDriver()
        self._disable_background_threads(driver)

        driver._my_pid = None
        driver._post_fork_init(voting=True)
        parent_elector = driver.elector

        # Simulate a worker re-initializing in the same process object
        driver._my_pid = 99999
        driver._post_fork_init(voting=False)

        self.assertIs(driver.elector, parent_elector)
        m_elector.assert_called_once()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2025 Tigera, Inc. All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib.api.definitions import portbindings
from neutron_lib.db import constants as db_consts
from neutron_lib.services.qos import base
from neutron_lib.services.qos import constants as qos_consts

from oslo_log import log


LOG = log.getLogger(__name__)

DRIVER = None

SUPPORTED_RULES = {
    qos_consts.RULE_TYPE_BANDWIDTH_LIMIT: {
        qos_consts.MAX_KBPS: {"type:range": [0, db_consts.DB_INTEGER_MAX_VALUE]},
        qos_consts.MAX_BURST: {"type:range": [0, db_consts.DB_INTEGER_MAX_VALUE]},
        qos_consts.DIRECTION: {"type:values": constants.VALID_DIRECTIONS},
    },
    qos_consts.RULE_TYPE_PACKET_RATE_LIMIT: {
        qos_consts.MAX_KPPS: {"type:range": [0, db_consts.DB_INTEGER_MAX_VALUE]},
        qos_consts.MAX_BURST_KPPS: {"type:range": [0, 0]},
        qos_consts.DIRECTION: {"type:values": constants.VALID_DIRECTIONS},
    },
}


class CalicoQoSDriver(base.DriverBase):

    @staticmethod
    def create(mechanism_driver):
        qd = CalicoQoSDriver(
            name="calico",
            vif_types=[portbindings.VIF_TYPE_TAP],
            vnic_types=[portbindings.VNIC_NORMAL],
            supported_rules=SUPPORTED_RULES,
            requires_rpc_notifications=False,
        )
        qd.mechanism_driver = mechanism_driver
        return qd

    def update_policy(self, context, policy):
        """Update policy invocation.

        This method can be implemented by the specific driver subclass
        to update the backend where necessary.

        :param context: current running context information
        :param policy: a QoSPolicy object being updated.
        """
        LOG.info("update_policy: context=%r policy=%r", context, policy)
        self.mechanism_driver.handle_qos_policy_update(context, policy.id)


def register(mechanism_driver):
    """Register the driver."""
    global DRIVER
    if not DRIVER:
        DRIVER = CalicoQoSDriver.create(mechanism_driver)
    LOG.debug("Calico QoS driver registered")



#!/usr/bin/env python3

# NOTE: This script makes optional use of some modules for 'nicer'
#       output, and I highly recommend installing them if you're
#       using this script manually, or testing it; however, the
#       entire script can run using only modules in the Python 3
#       standard library, so that no docker container, requirements.txt,
#       or virtualenv is required.
#
#       We also support inline dependencies a la PEP 723, and recommend
#       using `uv` for this project (and every other project):
#
#           https://docs.astral.sh/uv/

"""
This script takes an expiry time and an image name on
quay.io and uses the API to set the expiry on that image.
"""

# /// script
# requires-python = ">=3.9"
# dependencies = [
#     "rich",
#     "rich-argparse",
# ]
# ///

import os
import sys
import json
import logging
import argparse
import datetime

from functools import cache
from enum import IntEnum, auto

from typing import Optional

import urllib.error
import urllib.request

try:
    # Use nicer printing if we have it
    from rich import print
except ImportError:
    pass

try:
    # Use nicer help formatting if we have it
    from rich_argparse.contrib import ParagraphRichHelpFormatter as HelpFormatter
except ImportError:
    from argparse import HelpFormatter

MONO_FORMAT = "[%(levelname)-8s] %(asctime)s [%(funcName)s:%(lineno)d] %(message)s"

HELP_DESCRIPTION_TEXT = """
A simple script to add an expiration date to an image tag on quay.io, or to remove the expiration date from one, using the quay.io API.

When removing an expiry it's also possible to un-expire an image thanks to some peculiarities of the quay.io tag API, since the API returns the list of previous tag data for a specific tag even if it's expired.
"""

HELP_EPILOGUE_TEXT = """
For prettier logging output, `pip install rich`. For prettier --help output, `pip install rich_argparse`.
"""


try:
    # Use nicer log formatting if we have it
    from rich.logging import RichHandler

    handler = RichHandler(rich_tracebacks=True)
except ImportError:
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(MONO_FORMAT))

handler.setLevel(logging.INFO)
logging.basicConfig(
    datefmt=r"%Y-%m-%d %H:%M:%S",
    level=logging.DEBUG,
    format="%(message)s",
    handlers=[handler],
)

log = logging.getLogger(__name__)

GET_TAG_INFO_URL = (
    "https://quay.io/api/v1/repository/{image_name}/tag/?specificTag={image_tag}"
)
PUT_TAG_INFO_URL = "https://quay.io/api/v1/repository/{image_name}/tag/{image_tag}"
DATETIME_FORMAT = "%a, %d %b %Y %X %z"


class ReturnCodes(IntEnum):
    """
    common return codes used in this module
    """

    OK = 0
    NOAPITOKEN = auto()
    INVALIDIMAGENAME = auto()
    HTTPGETTAGINFOFAILED = auto()
    HTTPSETEXPIRYFORBIDDEN = auto()
    HTTPIMAGENOTFOUND = auto()


class Image:
    """
    The Image class represents a docker image on Quay.io, and
    includes functionality to interact with its API to get information
    and set expiry times.
    """

    def __init__(self, image_identifier):
        if image_identifier.count(":") > 1:
            log.critical(
                "Image name needs to be of the format quay.io/image/name:<tag>"
            )
            sys.exit(ReturnCodes.INVALIDIMAGENAME)
        image_uri, image_tag = image_identifier.split(":")
        self.image_identifier = image_identifier
        self.image_uri = image_uri
        self.image_name = image_uri.removeprefix("quay.io/")
        self.image_tag = image_tag

        self.data = None

        self.fetch_tag_info()
        self.validate_dates()

    @property
    def last_modified_datetime(self):
        return datetime.datetime.strptime(
            self.latest_tag["last_modified"], DATETIME_FORMAT
        )

    @property
    def has_expiry(self):
        return self.expiry_datetime is not None

    @property
    def tag_data(self):
        if self.data is None:
            self.fetch_tag_info()
        return self.data

    @property
    def latest_tag(self):
        return self.data["tags"][0]

    @property
    def expiry_datetime(self):
        try:
            return datetime.datetime.strptime(
                self.latest_tag["expiration"], DATETIME_FORMAT
            )
        except KeyError:
            return None

    @property
    def is_expired(self):
        if self.expiry_datetime:
            return self.expiry_datetime < datetime.datetime.now(tz=datetime.timezone.utc)
        else:
            return False

    def fetch_tag_info(self):
        log.debug(
            "Fetching info for image tag '%s' from quay.io API", self.image_identifier
        )
        req = urllib.request.Request(
            url=GET_TAG_INFO_URL.format(
                image_name=self.image_name, image_tag=self.image_tag
            ),
            headers=auth_header(),
            method="GET",
        )
        log.debug("Fetching tag data for tag '%s'", self.image_identifier)
        try:
            with urllib.request.urlopen(req) as response:
                data = json.load(response)
                self.data = data
        except urllib.error.HTTPError as http_exc:
            log.error(
                "Caught HTTP error getting tag data for '%s': %s",
                self.image_identifier,
                http_exc,
            )
            sys.exit(ReturnCodes.HTTPGETTAGINFOFAILED)


    def validate_dates(self):
        log.debug("Last modified date: '%s'", self.last_modified_datetime)
        log.debug("Expiry date: '%s'", self.last_modified_datetime)
        if self.expiry_datetime:
            if self.is_expired:
                log.error(
                    (
                        "Image tag '%s' exists in the API, but is already expired! "
                        "Further API calls against this tag are likely to fail."
                    ),
                    self.image_identifier,
                )
            else:
                log.debug(
                    "Image tag '%s' has configured expiration '%s'",
                    self.image_identifier,
                    self.expiry_datetime,
                )
        else:
            log.debug(
                "Image tag '%s' has no configured expiration", self.image_identifier
            )

    def __put_tag_info(self, info):
        req = urllib.request.Request(
            url=PUT_TAG_INFO_URL.format(
                image_name=self.image_name, image_tag=self.image_tag
            ),
            headers=auth_header(),
            data=json.dumps(info).encode(),
            method="PUT",
        )
        req.add_header("Content-Type", "application/json")
        try:
            with urllib.request.urlopen(req) as response:
                if response.code == 201:
                    log.debug(
                        "Successfully updated tag '%s' with data %s",
                        self.image_identifier,
                        info,
                    )
                else:
                    log.warning("Got unexpected HTTP response %s", response.status)
        except urllib.error.HTTPError as http_exc:
            log.error(
                "Caught HTTP error setting tag data for '%s': %s",
                self.image_identifier,
                http_exc,
            )

    def remove_expiry(self, restore=False):
        """
        Removes the expiry date from an image. Due to a quirk of
        the API, we can also use this functionality to restore an
        image that has already expired.

        Args:
            restore (bool, optional): Whether or not to restore the image if it's expired. Defaults to False.
        """

        # If the current tag doesn't have an expiry, that means there's nothing
        # to remove and we don't need to restore it even if we were asked to.
        if self.has_expiry is False:
            log.info(
                "Image tag '%s' has no expiry set, skipping", self.image_identifier
            )
            return

        # Setting only the manifest digest creates a "new" entry in the tag's history,
        # which, by default, has no expiry set. This not only means we can use this
        # to remove expiry, but we can also use it to restore an already-expired image
        # to its previous, un-expired version.
        data = {"manifest_digest": self.latest_tag["manifest_digest"]}

        # Image is expired and we're not trying to restore it - do nothing.
        if self.is_expired and restore is False:
            log.error(
                "Image tag '%s' has expired, skipping (use --restore to restore this image)",
                self.image_identifier,
            )
            return

        # Here we set the tag's manifest_digest but nothing else
        self.__put_tag_info(data)

        # Do some logic to determine what our log should be like
        if self.is_expired:
            log.info(
                "Restored expired image tag '%s' to manifest '%s' from '%s'",
                self.image_identifier,
                self.latest_tag["manifest_digest"],
                self.last_modified_datetime,
            )
        elif self.has_expiry:
            log.info(
                "Removed expiry '%s' from image tag '%s'",
                self.expiry_datetime,
                self.image_identifier,
            )

    def set_expiry(self, days: int, relative=False, override=False):
        if self.has_expiry:
            if override is False:
                log.warning(
                    "Image tag '%s' already has expiry set, skipping",
                    self.image_identifier,
                )
                return
            log.warning(
                "Image tag '%s' already has expiry set, but --force-override was set, proceeding anyway",
                self.image_identifier,
            )

        if relative:
            expiry_datetime = make_expiry_datetime(days, self.last_modified_datetime)
        else:
            expiry_datetime = make_expiry_datetime(days)

        data = {"expiration": expiry_datetime.timestamp()}
        self.__put_tag_info(data)
        log.info(
            "Set expiry information for image tag '%s' to '%s'",
            self.image_identifier,
            expiry_datetime,
        )


@cache
def auth_header():
    return {"Authorization": f"Bearer {get_quay_token()}"}


def get_quay_token():
    try:
        token = os.environ["QUAY_API_TOKEN"]
        if not token:
            raise RuntimeError("QUAY_API_TOKEN is unset or empty")
    except (KeyError, RuntimeError):
        log.critical(
            "Please set QUAY_API_TOKEN in the environment before running this script"
        )
        sys.exit(ReturnCodes.NOAPITOKEN)
    return token


def make_expiry_datetime(days: int, starting_date: Optional[datetime.datetime] = None):
    # Create a timedelta from the number of days passed in
    time_delta = datetime.timedelta(days=days)

    # Get the current timestamp. We'll use this as a default starting_date, but
    # also use it to validate that our expiration date is in the future.
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    if starting_date is None:
        starting_date = now

    # Strip away the time from the datetime so that we're operating at midnight
    # UTC (which makes things predictable to some extent)
    expiry_datetime = (
        starting_date.replace(hour=0, minute=0, second=0, microsecond=0) + time_delta
    )

    # The API won't let us set dates in the past, so we need to set the expiry
    # to be 'now' (or, to account for time skew, ten seconds from now)
    if now > expiry_datetime:
        log.warning(
            "Target expiry '%s' is in the past! Setting expiry to ten seconds from now instead.",
            expiry_datetime,
        )
        expiry_datetime = now + datetime.timedelta(minutes=10)
    return expiry_datetime


def main():
    parser = argparse.ArgumentParser(
        formatter_class=HelpFormatter,
        description=HELP_DESCRIPTION_TEXT,
        epilog=HELP_EPILOGUE_TEXT,
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="Enable debug logging"
    )

    subparsers = parser.add_subparsers(dest="action")
    parser_add_expiry = subparsers.add_parser(
        "add", help="Add expiry to an image", formatter_class=parser.formatter_class
    )

    parser_add_expiry.add_argument(
        "image_names", nargs="+", help="The image to add expiry to"
    )
    parser_add_expiry.add_argument(
        "--expiry-days",
        type=int,
        default=90,
        help="The time (in days from today UTC) to expire the images (default: 90d)",
    )
    parser_add_expiry.add_argument(
        "--relative-date",
        action="store_true",
        default=False,
        help="Set the expiry to days from last image modification instead",
    )
    parser_add_expiry.add_argument(
        "--force-override",
        action="store_true",
        default=False,
        help="Set expiry even if the image has an expiry set already",
    )

    parser_remove_expiry = subparsers.add_parser(
        "remove",
        help="Remove expiry to an image (and optionally un-expire the image)",
        formatter_class=parser.formatter_class,
    )
    parser_remove_expiry.add_argument(
        "image_names", nargs="+", help="The image to remove expiry from"
    )
    parser_remove_expiry.add_argument(
        "--restore",
        action="store_true",
        default=False,
        help="If the image tag is already expired, restore it to the last known tag",
    )

    args = parser.parse_args()

    if args.debug:
        handler.setLevel(logging.DEBUG)
        log.debug("Executing with args %s", args)

    for image_spec in args.image_names:
        image_spec = image_spec.removeprefix("https://")
        if not image_spec.startswith("quay.io"):
            log.error(
                "Image '%s' does not appear to be a quay.io image, skipping", image_spec
            )
            continue
        image_obj = Image(image_spec)
        if args.action == "add":
            image_obj.set_expiry(args.expiry_days, relative=args.relative_date, override=args.force_override)
        elif args.action == "remove":
            image_obj.remove_expiry(restore=args.restore)


if __name__ == "__main__":
    main()



# Copyright 2019 Tigera, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from etcd3gw.exceptions import Etcd3Exception

import mock

from neutron.tests import base

from oslo_log import log

from networking_calico import etcdv3


LOG = logging.getLogger(__name__)


class TestEtcdv3(base.BaseTestCase):

    def test_exception_detail_logging(self):

        e3e = Etcd3Exception(detail_text="from test_exception_detail_logging")

        def fn(self, *args, **kwargs):
            raise e3e

        wrapped = etcdv3.logging_exceptions(fn)

        with mock.patch.object(
            log.getLogger("networking_calico.etcdv3"), "warning"
        ) as mock_lw:
            try:
                wrapped(None)
            except Exception:
                pass
            mock_lw.assert_called_with(
                "Etcd3Exception, re-raising: %r:\n%s",
                e3e,
                "from test_exception_detail_logging",
            )



# Copyright (c) 2022 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import copy
import os

from tests.st.test_base import TestBase
from tests.st.utils.utils import log_and_run, calicoctl, \
    API_VERSION, name, namespace, ERROR_CONFLICT, NOT_FOUND, NOT_NAMESPACED, \
    SET_DEFAULT, NOT_SUPPORTED, KUBERNETES_NP, NOT_LOCKED_SPLIT, \
    POOL_NOT_EXIST_CIDR, INVALID_SPLIT_NUM, POOL_TOO_SMALL, \
    NOT_KUBERNETES, NO_IPAM, writeyaml
from tests.st.utils.data import *

logging.basicConfig(level=logging.DEBUG, format="%(message)s")
logger = logging.getLogger(__name__)

class TestCalicoctlIPAMSplit(TestBase):
    """
    Test splitting IPAM pools works
    1) Test splitting an ippool 10.0.1.0/24 into 4 ippools:
       10.0.1.128/26, 10.0.1.64/26, 10.0.1.0/26, and 10.0.1.192/26.
    """

    def test_ipam_split_by_cidr(self):
        """
        Test that splitting IP pools works properly by CIDR
        """

        # Create the ipv4 pool using calicoctl, and read it out using an
        # exact get and a list query.
        rc = calicoctl("create", data=ippool_name1_rev1_v4)
        rc.assert_no_error()
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_v4))
        rc.assert_data(ippool_name1_rev1_v4)
        rc = calicoctl("get ippool -o yaml")
        rc.assert_list("IPPool", [ippool_name1_rev1_v4])

        # Create a Node, this should also trigger auto-creation of a cluster info
        rc = calicoctl("create", data=node_name4_rev1)
        rc.assert_no_error()
        rc = calicoctl("get node %s -o yaml" % name(node_name4_rev1))
        rc.assert_data(node_name4_rev1)
        rc = calicoctl("get clusterinfo %s -o yaml" % name(clusterinfo_name1_rev1))
        rc.assert_no_error()

        # Attempt to split the IP pool before locking the datastore
        rc = calicoctl("ipam split --cidr=10.0.1.0/24 4")
        rc.assert_error(text=NOT_LOCKED_SPLIT)

        # Lock the data
        rc = calicoctl("datastore migrate lock")
        rc.assert_no_error()

        # Attempt to split a nonexistent IP pool
        rc = calicoctl("ipam split --cidr=10.0.2.0/24 4")
        rc.assert_error(text=POOL_NOT_EXIST_CIDR)

        # Attempt to split an IP pool into an invalid number of child pools
        rc = calicoctl("ipam split --cidr=10.0.1.0/24 3")
        rc.assert_error(text=INVALID_SPLIT_NUM)

        # Attempt to split an IP pool into more pools than possible given the size
        rc = calicoctl("ipam split --cidr=10.0.1.0/24 512")
        rc.assert_error(text=POOL_TOO_SMALL)

        # Split the IP pool
        rc = calicoctl("ipam split --cidr=10.0.1.0/24 4")
        rc.assert_no_error()

        # Check that the original IP pool no longer exists
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_v4))
        rc.assert_error(text=NOT_FOUND)

        # Check that the split IP pools exist
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split1_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split1_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split2_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split2_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split3_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split3_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split4_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split4_v4)

        # Unlock the datastore
        rc = calicoctl("datastore migrate unlock")
        rc.assert_no_error()

        # Clean up
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split1_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split2_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split3_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split4_v4))
        rc.assert_no_error()
        rc = calicoctl("delete node %s" % name(node_name4_rev1))
        rc.assert_no_error()

    def test_ipam_split_by_name(self):
        """
        Test that splitting IP pools works properly by IP pool name
        """

        # Create the ipv4 pool using calicoctl, and read it out using an
        # exact get and a list query.
        rc = calicoctl("create", data=ippool_name1_rev1_v4)
        rc.assert_no_error()
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_v4))
        rc.assert_data(ippool_name1_rev1_v4)
        rc = calicoctl("get ippool -o yaml")
        rc.assert_list("IPPool", [ippool_name1_rev1_v4])

        # Create a Node, this should also trigger auto-creation of a cluster info
        rc = calicoctl("create", data=node_name4_rev1)
        rc.assert_no_error()
        rc = calicoctl("get node %s -o yaml" % name(node_name4_rev1))
        rc.assert_data(node_name4_rev1)
        rc = calicoctl("get clusterinfo %s -o yaml" % name(clusterinfo_name1_rev1))
        rc.assert_no_error()

        # Attempt to split the IP pool before locking the datastore
        rc = calicoctl("ipam split --name=%s 4" % name(ippool_name1_rev1_v4))
        rc.assert_error(text=NOT_LOCKED_SPLIT)

        # Lock the data
        rc = calicoctl("datastore migrate lock")
        rc.assert_no_error()

        # Attempt to split a nonexistent IP pool
        rc = calicoctl("ipam split --name=%s 4" % name(ippool_name2_rev1_v6))
        rc.assert_error(text=POOL_NOT_EXIST_CIDR)

        # Attempt to split an IP pool into an invalid number of child pools
        rc = calicoctl("ipam split --name=%s 3" % name(ippool_name1_rev1_v4))
        rc.assert_error(text=INVALID_SPLIT_NUM)

        # Attempt to split an IP pool into more pools than possible given the size
        rc = calicoctl("ipam split --name=%s 512" % name(ippool_name1_rev1_v4))
        rc.assert_error(text=POOL_TOO_SMALL)

        # Split the IP pool
        rc = calicoctl("ipam split --name=%s 4" % name(ippool_name1_rev1_v4))
        rc.assert_no_error()

        # Check that the original IP pool no longer exists
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_v4))
        rc.assert_error(text=NOT_FOUND)

        # Check that the split IP pools exist
        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split1_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split1_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split2_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split2_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split3_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split3_v4)

        rc = calicoctl("get ippool %s -o yaml" % name(ippool_name1_rev1_split4_v4))
        rc.assert_no_error()
        rc.assert_data(ippool_name1_rev1_split4_v4)

        # Unlock the datastore
        rc = calicoctl("datastore migrate unlock")
        rc.assert_no_error()

        # Clean up
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split1_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split2_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split3_v4))
        rc.assert_no_error()
        rc = calicoctl("delete ippool %s" % name(ippool_name1_rev1_split4_v4))
        rc.assert_no_error()
        rc = calicoctl("delete node %s" % name(node_name4_rev1))
        rc.assert_no_error()
